""" 
GraphColorFlow Library (Version 0.0.2)
Copyright 2022 Linguistic Agents. All Rights Reserved.
Licensed under the MIT License
requires ClassIdGraph
"""

import numpy as np
import tensorflow as tf
from tensorflow.data import Dataset

import tensorflow.keras as keras
from keras.layers import Dense, Input
from keras.models import Model

from classidgraph.classidgraph  import Graph, Node, NodeLevel

from typing import List
Vector = List[float]

class AutoencodersForOneLevel:
    """ Group of Autoencoders for all colors from the same Round
    """
    def __init__(self, level : int, num_colors : int, num_inputs : int, num_dimensions : int, learning_rate : float):
        super(AutoencodersForOneLevel, self).__init__()
        self.level = level
        self.autoencoders : List[Autoencoder] = []

        for _ in range(num_colors):
            autoencoder = Autoencoder(num_colors = num_colors, num_inputs_per_color = num_inputs, num_dimensions = num_dimensions)
            autoencoder.compile (   
                                    optimizer=tf.keras.optimizers.Adam(learning_rate=learning_rate), 
                                    loss=tf.keras.losses.MeanSquaredError()
                                )
            self.autoencoders.append(autoencoder)

    def train_for_one_color(self, color : int, x : List[Vector], y : List[Vector], num_epochs : int, verbose = False):
        autoencoder = self.autoencoders[color]

        #the_optinizer = tf.keras.optimizers.Adam(learning_rate=0.00001)
        #the_loss = tf.keras.losses.MeanSquaredError()
        #the_metric : tf.keras.metrics.Metric = tf.keras.metrics.MeanSquaredError()

        # train loop
        #for n_epoch in range(num_epochs):
        autoencoder.fit(x = x, y = y, validation_split=0.2, epochs=num_epochs, batch_size=64, verbose = verbose)    

    def encode(self, color : int, input_vector : Vector):
        autoencoder = self.autoencoders[color]
        z = autoencoder.encoder.predict( np.array([ input_vector ]) )

        return  z

class Autoencoder(Model):
    """ Group of Autoencoders for all colors from the same Round
    """

    def __init__(self, num_colors : int, num_inputs_per_color : int, num_dimensions : int):
        super(Autoencoder, self).__init__()

        self.encoder = tf.keras.Sequential([
            Input( (num_colors + 1) * num_inputs_per_color),
            Dense(units = num_dimensions, activation="tanh"), 
            Dense(units = num_dimensions, activation="tanh"),
            Dense(units = num_dimensions, activation="tanh"),
            Dense(units = num_dimensions, activation="tanh"),

            Dense(units = num_dimensions, activation="linear")
        ])   

        self.decoder = tf.keras.Sequential([
            Dense(num_inputs_per_color, activation='linear'),
        ])

    def call(self, x):
        y = x
        encoded = self.encoder(x)
        decoded = self.decoder(encoded)
        return  decoded

class NetworkParams:
    """ Hyperparameters defining a Graph Neural Network 
    """
    def __init__(self, num_levels : int, num_colors : int, num_dimensions : int):
        super(NetworkParams, self).__init__()
        self.num_levels = num_levels
        self.num_colors = num_colors
        self.num_dimensions = num_dimensions

class GraphNetwork:
    def __init__(self, graph : Graph, num_levels : int, num_colors : int, num_dimensions : int, learning_rate : float = 0.00001):
        self.graph = graph
        self.network_params = NetworkParams(num_levels = num_levels, num_colors = num_colors, num_dimensions = num_dimensions)
        self.levels = {}

        for level in range(self.network_params.num_levels):
            self.levels[level] = AutoencodersForOneLevel(level = level,
                num_colors = self.network_params.num_colors, 
                num_inputs = self.get_num_inputs(level = level),
                num_dimensions = self.network_params.num_dimensions,
                learning_rate = learning_rate)


    def  get_num_inputs(self, level : int):
        " Return the number of inputs in the given level of the network. "
        if level == 0:
            return  self.graph.num_classes + self.graph.num_features
        else:
            return  self.network_params.num_dimensions

    def train(self, epochs : int, verbose = False):
        all_nodes = self.graph.get_all_nodes()

        for node in all_nodes:
            self.initialize_level_0_for_node(node = node)

        for level in self.levels:
            # aggregate data from sources and goals 
            self.graph.aggregate_data(level=level)           
            # train on one level for all nodes
            self.train_level(level = level, num_epochs = epochs, verbose = verbose)

    def initialize_level_0_for_node(self, node : Node):
        class_vector = self.graph.class_vectors[node.class_]

        # combine Class Vector and Feature Vector
        the_list = [ *class_vector, *node.features ]
        combined_vector : Vector = np.array(the_list)

        node_level = NodeLevel(node = node, level = 0, num_vectors = self.network_params.num_colors, num_elements = len(combined_vector) )  

        for color in range(self.network_params.num_colors):
            node_level. incoming.vector_list[color] = combined_vector.copy()
            node_level.intrinsic.vector_list[color] = combined_vector.copy()
            node_level. outgoing.vector_list[color] = combined_vector.copy()

        node.all_levels.append(node_level)

    def train_level(self, level : int, num_epochs : int, verbose = False):
        num_levels : int = len(self.levels)
        all_nodes = self.graph.get_all_nodes()

        autoencoders : AutoencodersForOneLevel = self.levels[level]

        if level >= num_levels - 1:
            return

        for color in range(self.network_params.num_colors):  # for each color 
            # prepare the samples for this color
            oX, oY = self.to_sample_sets(level = level, color = color)

            # train the encoder-decoder network 
            # at this Round
            autoencoders.train_for_one_color(color = color, x = oX, y = oY, num_epochs = num_epochs, verbose = verbose)
            
        for node in all_nodes:   # at this level
            assert  len(node.all_levels) == level + 1   # to be ready for adding the next level
            next_node_level = self.process_one_node(level = level, node = node)
            node.all_levels.append(next_node_level)

    def process_one_node(self, level : int, node : Node):
        autoencoders : AutoencodersForOneLevel = self.levels[level]

        for_all_colors : List[Vector] = []
        for color in range(self.network_params.num_colors):  # for each color 
            """ 
            all the aggregate sums already calculated, 
            a new 'intrinsic' vector is calculated using the 'encoder' part of the autoencoder 
            """
            input_vector : Vector = node.get_input_vector(level = level, n_vector = color)
            encoded : Vector = autoencoders.encode(color = color, input_vector = input_vector)
            for_all_colors.append(encoded[0])    

        next_node_level = NodeLevel(node, level, self.network_params.num_colors, self.network_params.num_dimensions)
        # set the 'intrinsic' vector of the next level
        next_node_level.set_center(for_all_vectors = for_all_colors)

        return  next_node_level

    def to_sample_sets(self, level : int, color : int):
        assert  len(self.graph.nodes_by_class) > 0

        input_vectors = []
        output_vectors = []

        for node in self.graph.get_all_nodes():
            input_vector = node.get_input_vector(level = level, n_vector = color)
            input_vectors.append(input_vector)

            output_vector = node.get_output_vector(level = level, n_vector = color)
            output_vectors.append(output_vector)

        return  (np.array(input_vectors), np.array(output_vectors)) 

